"""
Classes to store features mapping for bucketing.
"""

import dataclasses
from dataclasses import dataclass, field
from typing import Dict, List, Optional, Union

import numpy as np
import pandas as pd


@dataclass
class BucketMapping:
    """Internal class to store all the bucketing info.

    This saves the information from a fitted bucketer and can transform any new data.

    Example:

    ```python
    from skorecard.bucket_mapping import BucketMapping

    # Manually a new bucket mapping for a feature
    bucket = BucketMapping('feature1', 'numerical', map = [2,3,4,5], specials={"special 0": [0]})
    print(bucket)

    # You can work with these classes as dicts as well
    bucket.as_dict()
    BucketMapping(**bucket.as_dict())

    # Transform new data
    bucket.transform(list(range(10)))
    ```

    Args:
        feature_name (str): Name of the feature
        type (str): Type of feature, one of ['categorical','numerical']
        missing_treatment (int): How missing values should be treated.
            If none, missing values get put in their separate bucket.
            If an int, this bucket number is where we will put the missing values.
        other_bucket (int):  Only for categoricals: If specified,
            the bucket number where any new values (not in the map) should be put.
        map (list or dict): The info needed to create the buckets (boundaries or cats)
        right (bool): parameter to np.digitize, used when map='numerical'.
        specials (dict): dictionary of special values to bin separately.
            The key is used for the labels, the value(s) is used to create a separate bucket.
        labels (dict): dictionary containing special values. It must be of the format:
            - keys: strings, containing the name (that will be used as labels) for the special values
            - values: lists, containing the special values
    """

    feature_name: str
    type: str
    missing_bucket: Optional[int] = None
    other_bucket: Optional[int] = None
    map: Union[Dict, List] = field(default_factory=lambda: [])
    right: bool = True
    specials: Dict[str, list] = field(default_factory=lambda: {})

    def __post_init__(self) -> None:
        """
        Do input validation, pre-calculations and build labels.
        """
        # Input validation
        assert self.type in ["numerical", "categorical"]
        assert len(self.map) is not None, "Please set a 'map' first"
        assert isinstance(self.specials, dict) or isinstance(self.specials, list)

        # Check specials
        assert all([isinstance(k, str) for k in self.specials.keys()]), f"The keys of the special dictionary must be \
        strings, got {self.specials.keys()} instead."
        assert all(
            [isinstance(k, list) for k in self.specials.values()]
        ), f"The values of the special dictionary must be a list of elements, got {self.specials}instead."
        # TODO: assert that special values are not present in multiple special buckets.

        # Make sure map is in correct format
        if isinstance(self.map, np.ndarray):
            self.map = self.map.tolist()

        # Determine the bucket numbers for reserved categories: 'missing'.
        # Missings are always -1 unless specified
        if self.type == "numerical":
            assert isinstance(self.map, list), "Map must be list"

            # 2 boundaries will give 3 buckets.
            # they are zero-indexed so max_bucket number equals len(self.map)
            max_bucket = len(self.map)

            if self.missing_bucket is not None:
                if not np.isnan(self.missing_bucket):
                    assert (
                        self.missing_bucket <= max_bucket
                    ), "map '{}' corresponds buckets 0-{} but missing_bucket is set to {}".format(
                        self.map,
                        max_bucket,
                        self.missing_bucket,
                    )
                self._missing_bucket = self.missing_bucket
            else:
                self._missing_bucket = -1

            # We leave -2 for 'other', which only applies to categoricals
            # There for consistency
            self._start_special_bucket = -3

            # Build labels
            self.labels = build_labels(
                self.map,
                right=self.right,
                missing_bucket=self._missing_bucket,
                specials=self.specials,
                start_special_bucket=self._start_special_bucket,
            )

        # Determine the bucket numbers for reserved categories: 'other' and 'missing'.
        if self.type == "categorical":
            assert isinstance(self.map, dict), "Map must be dict"

            # Assure the conversion from numpy.int to int
            new_dict = dict()
            for k, v in self.map.items():
                if isinstance(k, (np.int32, np.int64)):
                    k = int(k)
                if isinstance(v, (np.int32, np.int64)):
                    v = int(v)
                new_dict[k] = v

            self.map = new_dict
            self._validate_categorical_map()
            self.map = dict(self.map)  # type: Dict

            # Python 3.7+ has dicts that are OrderedDicts. Let's have a pretty ordering
            self.map = dict(sorted(self.map.items(), key=lambda x: (x[1], x[0])))

            # Set 'other' bucket
            if self.other_bucket is not None:
                assert (
                    self.other_bucket in self.map.values()
                ), "other_bucket '{}' does not exist in map values: {}".format(
                    self.other_bucket,
                    self.map,
                )
                self._other_bucket = self.other_bucket
            else:
                self._other_bucket = -2

            # Set 'missing' bucket
            if self.missing_bucket is not None:
                # Allow -2, -1 Some missing_treatments (e.g. most_risky) add it here
                if self.missing_bucket not in [-2, -1, np.nan]:
                    assert (
                        self.missing_bucket in self.map.values()
                    ), f"missing_bucket '{self.missing_bucket}' does not exist in map values: {self.map}"

                self._missing_bucket = self.missing_bucket
            else:
                self._missing_bucket = -1

            # Set specials bucket
            self._start_special_bucket = -3

            # Build labels
            self.labels = build_cat_labels(
                self.map,
                missing_bucket=self._missing_bucket,
                other_bucket=self._other_bucket,
                specials=self.specials,
                start_special_bucket=self._start_special_bucket,
            )

    def _validate_categorical_map(self):
        """
        Validate map structure.

        Assures that the provided mapping starts at 0 and that it has an incremental trend.
        """
        values = [v for v in self.map.values()]
        if len(values) > 0:
            if not np.array_equal(np.unique(values), np.arange(max(values) + 1)):
                err_msg = (
                    f"Mapping dictionary must start at 0 and be incremental. "
                    f"Found the following mappings {np.unique(values)}, and expected {np.arange(max(values) + 1)}"
                )
                raise ValueError(err_msg)

    def transform(self, x):
        """
        Apply binning using a boundaries map.

        Note:
        - resulting bins are zero-indexed

        ```python
        import numpy as np
        bins = np.array([-np.inf, 1, np.inf])
        x = np.array([-1,0,.5, 1, 1.5, 2,10, np.nan, 0])
        new = np.digitize(x, bins)
        np.where(np.isnan(x), np.nan, new)
        ```
        """
        if isinstance(x, np.ndarray):
            x = pd.Series(x)
        if isinstance(x, list):
            x = pd.Series(x)
        assert isinstance(x, pd.core.series.Series)

        # Workaround for missings
        def to_int(x):
            if not np.isnan(x):
                return int(x)
            else:
                return x

        # Transform using self.map
        if self.type == "numerical":
            buckets = np.digitize(x, self.map, right=self.right).astype(int)
            buckets = pd.Series(buckets)
        if self.type == "categorical":
            buckets = self._apply_cat_mapping(x)

        # Deal with missing values
        buckets = np.where(x.isnull(), self._missing_bucket, buckets)

        # Ensure dtype is integer buckets
        buckets = [to_int(x) for x in buckets]

        # Deal with special values
        # Both categorical & numerical
        special_counter = self._start_special_bucket
        for k, v in self.specials.items():
            buckets = np.where(x.isin(v), special_counter, buckets)
            special_counter -= 1
        return np.array(buckets)

    def _apply_cat_mapping(self, x):
        mapping = MissingDict(self.map)
        mapping.set_missing_value(self._other_bucket)  # This was 'other' but you cannot mix integers and strings

        # Note that we need to add .astype('Int64')
        # so that we have a nullable integer series.
        # This is because na_action='ignore' will convert from Int64 series to float64
        # Support for nullable integpyer arrays is a pandas 'gotcha'
        # Details discussed here:
        # https://pandas.pydata.org/pandas-docs/stable/user_guide/gotchas.html#support-for-integer-na

        return x.map(mapping, na_action="ignore").astype("Int64")

    def as_dict(self) -> dict:
        """Return data in class as a dict.

        Returns:
            dict: data in class
        """
        return dataclasses.asdict(self)


class MissingDict(dict):
    """
    Deal with missing entries in a dict map.

    Because Pandas .map() uses the __missing__ method
    https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.Series.map.html

    Note the word 'missing' is confusing: It refers to the situation when a value is not
    in the dict map, and sets a default value instead.

    Missing values can be dealt with separately using map(na_action='ignore')

    Example usage:

    ```python
    s = pd.Series(['cat', 'dog', np.nan, 'rabbit'])
    a = {'cat': 'kitten', 'dog': 'puppy'}
    s.map(a)
    a = MissingDict(a)
    a.set_missing_value("bye")
    s.map(a)
    # make sure to ignore missing values!
    s.map(a, na_action='ignore')
    ```
    """

    def set_missing_value(self, value):
        """Setter for a missing value."""
        self.missing_value = value

    def __missing__(self, key):
        """Adds a default for missing values."""
        assert self.missing_value is not None, "Use .set_missing_value(key) first"
        return self.missing_value


def build_labels(
    boundaries,
    right: bool,
    missing_bucket: int,
    start_special_bucket: int,
    specials: Optional[Dict[str, list]],
) -> Dict[int, str]:
    """
    Build a nice label dict from a boundary.

    ```python
    assert build_labels([1,2,3]) == {
        0: '(-inf, 1.0]',
        1: '(1.0, 3.0]',
        2: '(3.0, 5.0]',
        3: '(5.0, inf]'
     }
    ```
    """
    boundaries = np.hstack([-np.inf, boundaries, np.inf]).tolist()
    labels = {}

    if right:
        b_left = "("
        b_right = "]"
    else:
        b_left = "["
        b_right = ")"

    for i, boundary in enumerate(boundaries):
        if i != len(boundaries) - 1:
            labels[i] = f"{b_left}{boundary}, {boundaries[i+1]}{b_right}"

    # reserve a label for missing values
    if missing_bucket in labels.keys():
        labels[missing_bucket] += " | Missing"
    else:
        labels[missing_bucket] = "Missing"

    # labels for specials
    if specials:
        for k, v in specials.items():
            labels[start_special_bucket] = "Special: " + str(k)
            start_special_bucket -= 1

    return labels


def build_cat_labels(
    boundaries,
    other_bucket: int,
    missing_bucket: int,
    start_special_bucket: int,
    specials: Optional[Dict[str, list]],
) -> Dict[int, str]:
    """
    Build categorical labels.
    """
    assert other_bucket is not None
    assert missing_bucket is not None

    # Switch value and key
    labels: Dict = {}
    for key in boundaries.keys():
        labels.setdefault(boundaries[key], []).append(str(key))

    # Merge lists into string
    for key in labels:
        labels[key] = ", ".join(labels[key])

    # add label for 'other'
    if other_bucket in labels.keys():
        labels[other_bucket] += " | Other"
    else:
        labels[other_bucket] = "Other"

    # reserve a label for missing values
    if missing_bucket in labels.keys():
        labels[missing_bucket] += " | Missing"
    else:
        labels[missing_bucket] = "Missing"

    # Add specials
    if specials:
        for k, v in specials.items():
            labels[start_special_bucket] = "Special: " + str(k)
            start_special_bucket -= 1

    # Make sure labels dict is nicely sorted by value, key
    labels = dict(sorted(labels.items(), key=lambda x: (x[1], x[0])))

    return labels


def merge_bucket_mapping(a, b):
    """
    Merges two bucketmappings into one.

    Assumption here is that one is for prebucketing and the other is for bucketing.
    In other words, one bucketmapping builds on the other one.
    """
    msg = f"Feature '{a.feature_name}' has variable_type '{a.type}' in a, but '{b.type}' in b."
    msg += "\nDid you set variable_type correctly in your (pre)bucketing pipeline?"
    assert a.type == b.type, msg

    if a.type == "categorical":
        if b.other_bucket:
            assert (
                b.other_bucket in a.labels.keys()
            ), f"b.other_bucket set to {b.other_bucket} but not present in any of a's buckets ({a.labels})"
        if b.missing_bucket:
            assert (
                b.missing_bucket in a.labels.keys()
            ), f"b.other_bucket set to {b.missing_bucket} but not present in any of a's buckets ({a.labels})"

        new_boundaries = {}
        for category, bucket in a.map.items():
            new_boundaries[category] = int(b.transform([bucket]))

        # let's also see where the 'other' category is assigned
        something_random = "84a088e251d2fa058f37145222e536dc"
        new_other_bucket = int(b.transform(a.transform([something_random])).tolist()[0])
        # if 'other' is put together with an existing bucket
        # manually assign that.
        if new_other_bucket in new_boundaries.values():
            other_bucket = new_other_bucket
        else:
            other_bucket = None

        # let's see where the missing category is assigned
        new_missing_bucket = int(b.transform(a.transform([np.nan])).tolist()[0])
        if new_missing_bucket in new_boundaries.values():
            missing_bucket = new_missing_bucket
        else:
            missing_bucket = None

        return BucketMapping(
            feature_name=a.feature_name,
            type=a.type,
            missing_bucket=missing_bucket,
            other_bucket=other_bucket,
            map=new_boundaries,
            specials=a.specials,
        )

    if a.type == "numerical":
        # This should hold for numerical maps
        assert len(a.map) >= len(b.map)

        # Add infinite edges to boundary map
        ref_map = [-np.inf] + a.map + [np.inf]

        new_buckets = list(b.transform(a.transform(ref_map)))

        # We take a.map and add inf edges, f.e.
        #  [-np.inf, 1,3,5, np.inf]
        # We can run it through both bucketers and get f.e.
        # [0,0,1,1,2]
        # If a.right = True, then we take the max per group
        # If a.right = False, then we take the min per group
        # Finally, remove any infinites
        new_boundaries = []

        if a.right:
            for i, new_bucket in enumerate(new_buckets):
                if i == 0:
                    if len(new_buckets) == 1:
                        new_boundaries.append(ref_map[0])
                        continue
                if i == len(new_buckets) - 1:
                    continue
                if new_buckets[i + 1] > new_bucket:
                    new_boundaries.append(ref_map[i])

        if not a.right:
            for i, new_bucket in enumerate(new_buckets):
                if i == 0:
                    if len(new_buckets) == 1:
                        new_boundaries.append(ref_map[0])
                        continue
                else:
                    if new_buckets[i - 1] < new_bucket:
                        new_boundaries.append(ref_map[i])

        new_boundaries = [x for x in new_boundaries if x != -np.inf]
        new_boundaries = [x for x in new_boundaries if x != np.inf]

        return BucketMapping(
            feature_name=a.feature_name,
            type=a.type,
            missing_bucket=a.missing_bucket,
            map=new_boundaries,
            specials=a.specials,
            right=a.right,
        )



import warnings
from typing import List

import numpy as np
import pandas as pd
import yaml
from sklearn.cluster import AgglomerativeClustering
from sklearn.tree import DecisionTreeClassifier, _tree
from sklearn.utils.multiclass import unique_labels

from skorecard.bucketers.base_bucketer import BaseBucketer
from skorecard.features_bucket_mapping import FeaturesBucketMapping
from skorecard.reporting import build_bucket_table
from skorecard.utils import NotInstalledError, NotPreBucketedError
from skorecard.utils.exceptions import ApproximationWarning
from skorecard.utils.validation import check_args, ensure_dataframe

try:
    from optbinning import OptimalBinning
except ModuleNotFoundError:
    OptimalBinning = NotInstalledError("optbinning")


class OptimalBucketer(BaseBucketer):
    """
    The `OptimalBucketer` transformer uses the [optbinning](http://gnpalencia.org/optbinning) package to find optimal buckets.

    Support: ![badge](https://img.shields.io/badge/numerical-true-green) ![badge](https://img.shields.io/badge/categorical-true-green) ![badge](https://img.shields.io/badge/supervised-true-green)

    This bucketer basically wraps optbinning.OptimalBinning to be consistent with skorecard.
    Requires a feature to be pre-bucketed to max 100 buckets.
    Optbinning uses a constrained programming solver to merge buckets,
    taking into account the following constraints 1) monotonicity in bad rate, 2) at least 5% of records per bin.

    Example:

    ```python
    from skorecard import datasets
    from skorecard.bucketers import OptimalBucketer

    X, y = datasets.load_uci_credit_card(return_X_y=True)
    bucketer = OptimalBucketer(variables = ['LIMIT_BAL'])
    bucketer.fit_transform(X, y)
    ```
    """  # noqa

    def __init__(
        self,
        variables=[],
        specials={},
        variables_type="numerical",
        max_n_bins=10,
        missing_treatment="separate",
        min_bin_size=0.05,
        cat_cutoff=None,
        time_limit=25,
        remainder="passthrough",
        get_statistics=True,
        solver="cp",
        monotonic_trend="auto_asc_desc",
        gamma=0,
        ob_kwargs={},
    ) -> None:
        """Initialize Optimal Bucketer.

        Args:
            variables: List of variables to bucket.
            specials: (nested) dictionary of special values that require their own binning.
                The dictionary has the following format:
                 {"<column name>" : {"name of special bucket" : <list with 1 or more values>}}
                For every feature that needs a special value, a dictionary must be passed as value.
                This dictionary contains a name of a bucket (key) and an array of unique values that should be put
                in that bucket.
                When special values are passed, they are not considered in the fitting procedure.
            variables_type: Passed to [optbinning.OptimalBinning](http://gnpalencia.org/optbinning/binning_binary.html):
                Type of the variables. Must be either 'categorical' or 'numerical'.
            missing_treatment: Defines how we treat the missing values present in the data.
                If a string, it must be one of the following options:
                    separate: Missing values get put in a separate 'Other' bucket: `-1`
                    most_risky: Missing values are put into the bucket containing the largest percentage of Class 1.
                    least_risky: Missing values are put into the bucket containing the largest percentage of Class 0.
                    most_frequent: Missing values are put into the most common bucket.
                    neutral: Missing values are put into the bucket with WoE closest to 0.
                    similar: Missing values are put into the bucket with WoE closest to the bucket with only missing values.
                    passthrough: Leaves missing values untouched.
                If a dict, it must be of the following format:
                    {"<column name>": <bucket_number>}
                    This bucket number is where we will put the missing values.
            min_bin_size: Passed to [optbinning.OptimalBinning](http://gnpalencia.org/optbinning/binning_binary.html):
                Minimum fraction of observations in a bucket.
            max_n_bins: Passed to [optbinning.OptimalBinning](http://gnpalencia.org/optbinning/binning_binary.html):
                Maximum numbers of bins to return.
            cat_cutoff: Passed to [optbinning.OptimalBinning](http://gnpalencia.org/optbinning/binning_binary.html):
                Threshold ratio (None, or >0 and <=1) below which categories are grouped
                together in a bucket 'other'.
            time_limit (float): Passed to [optbinning.OptimalBinning](http://gnpalencia.org/optbinning/binning_binary.html):
                Time limit in seconds to find an optimal solution.
            remainder: How we want the non-specified columns to be transformed. It must be in ["passthrough", "drop"].
                passthrough (Default): all columns that were not specified in "variables" will be passed through.
                drop: all remaining columns that were not specified in "variables" will be dropped.
            solver (str): Passed to [optbinning.OptimalBinning](http://gnpalencia.org/optbinning/binning_binary.html): The optimizer to solve the optimal binning problem.
                Supported solvers are “mip” to choose a mixed-integer programming solver, “cp” (default) to choose a constrained programming solver or “ls” to choose LocalSolver.
            monotonic_trend (str): Passed to [optbinning.OptimalBinning](http://gnpalencia.org/optbinning/binning_binary.html):
                The event rate monotonic trend. Supported trends are “auto”, “auto_heuristic” and “auto_asc_desc”
                to automatically determine the trend maximizing IV using a machine learning classifier,
                “ascending”, “descending”, “concave”, “convex”, “peak” and “peak_heuristic” to allow a peak change point,
                and “valley” and “valley_heuristic” to allow a valley change point.
                Trends “auto_heuristic”, “peak_heuristic” and “valley_heuristic” use a heuristic to determine the change point,
                and are significantly faster for large size instances (max_n_prebins > 20).
                Trend “auto_asc_desc” is used to automatically select the best monotonic trend
                between “ascending” and “descending”. If None, then the monotonic constraint is disabled.
            gamma (float): Passed to [optbinning.OptimalBinning](http://gnpalencia.org/optbinning/binning_binary.html):
                Regularization strength to reduce the number of dominating bins.
                Larger values specify stronger regularization. Default is 0.
                Option supported by solvers “cp” and “mip”.
            ob_kwargs (dict): Other parameters passed to [optbinning.OptimalBinning](http://gnpalencia.org/optbinning/binning_binary.html).
        """  # noqa
        self.variables = variables
        self.specials = specials
        self.variables_type = variables_type
        self.max_n_bins = max_n_bins
        self.missing_treatment = missing_treatment
        self.min_bin_size = min_bin_size
        self.cat_cutoff = cat_cutoff
        self.time_limit = time_limit
        self.remainder = remainder
        self.get_statistics = get_statistics
        self.solver = solver
        self.monotonic_trend = monotonic_trend
        self.gamma = gamma
        self.ob_kwargs = ob_kwargs

        check_args(ob_kwargs, OptimalBinning)

    def _get_feature_splits(self, feature, X, y, X_unfiltered=None):
        """
        Finds the splits for a single feature.

        X and y have already been preprocessed, and have specials removed.

        Args:
            feature (str): Name of the feature.
            X (pd.Series): df with single column of feature to bucket
            y (np.ndarray): array with target
            X_unfiltered (pd.Series): df with single column of feature to bucket before any filtering was applied

        Returns:
            splits, right (tuple): The splits (dict or array), and whether right=True or False.
        """
        # Normally Optbinning uses a DecisionTreeBucketer to do automatic prebinning
        # We require the user to pre-bucket explicitly before using this.
        if self.variables_type == "numerical":
            uniq_values = np.sort(np.unique(X.values))
            if len(uniq_values) > 100:
                raise NotPreBucketedError(
                    f"""
                    OptimalBucketer requires numerical feature '{feature}' to be pre-bucketed
                    to max 100 unique values (for performance reasons).
                    Currently there are {len(uniq_values)} unique values present.

                    Apply pre-binning, f.e. with skorecard.bucketers.DecisionTreeBucketer.
                    """
                )
            user_splits = uniq_values
        else:
            user_splits = None

        # Fit estimator
        binner = OptimalBinning(
            name=str(feature),
            dtype=self.variables_type,
            solver=self.solver,
            monotonic_trend=self.monotonic_trend,
            gamma=self.gamma,
            # On user_splits:
            # We want skorecard users to explicitly define pre-binning for numerical features
            # Setting the user_splits prevents OptimalBinning from doing pre-binning again.
            user_splits=user_splits,
            min_bin_size=self.min_bin_size,
            max_n_bins=self.max_n_bins,
            cat_cutoff=self.cat_cutoff,
            time_limit=self.time_limit,
            **self.ob_kwargs,
        )
        binner.fit(X.values, y)

        # Extract fitted boundaries
        if self.variables_type == "categorical":
            splits = {}
            for bucket_nr, values in enumerate(binner.splits):
                for value in values:
                    splits[value] = bucket_nr
        else:
            splits = binner.splits

        # Note that optbinning transform uses right=False
        # https://github.com/guillermo-navas-palencia/optbinning/blob/396b9bed97581094167c9eb4744c2fd1fb5c7408/optbinning/binning/transformations.py#L126-L132
        return (splits, False)


class EqualWidthBucketer(BaseBucketer):
    """
    The `EqualWidthBucketer` transformer creates equally spaced bins using [numpy.histogram](https://numpy.org/doc/stable/reference/generated/numpy.histogram.html)
    function.

    Support: ![badge](https://img.shields.io/badge/numerical-true-green) ![badge](https://img.shields.io/badge/categorical-false-red) ![badge](https://img.shields.io/badge/supervised-false-red)

    Example:

    ```python
    from skorecard import datasets
    from skorecard.bucketers import EqualWidthBucketer

    specials = {"LIMIT_BAL": {"=50000": [50000], "in [20001,30000]": [20000, 30000]}}

    X, y = datasets.load_uci_credit_card(return_X_y=True)
    bucketer = EqualWidthBucketer(n_bins = 10, variables = ['LIMIT_BAL'], specials=specials)
    bucketer.fit_transform(X)
    bucketer.fit_transform(X)['LIMIT_BAL'].value_counts()
    ```
    """  # noqa

    def __init__(
        self,
        n_bins=5,
        variables=[],
        specials={},
        missing_treatment="separate",
        remainder="passthrough",
        get_statistics=True,
    ):
        """Init the class.

        Args:
            n_bins (int): Number of bins to create.
            variables (list): The features to bucket. Uses all features if not defined.
            specials: (dict) of special values that require their own binning.
                The dictionary has the following format:
                 {"<column name>" : {"name of special bucket" : <list with 1 or more values>}}
                For every feature that needs a special value, a dictionary must be passed as value.
                This dictionary contains a name of a bucket (key) and an array of unique values that should be put
                in that bucket.
                When special values are defined, they are not considered in the fitting procedure.
            missing_treatment: Defines how we treat the missing values present in the data.
                If a string, it must be one of the following options:
                    separate: Missing values get put in a separate 'Other' bucket: `-1`
                    most_risky: Missing values are put into the bucket containing the largest percentage of Class 1.
                    least_risky: Missing values are put into the bucket containing the largest percentage of Class 0.
                    most_frequent: Missing values are put into the most common bucket.
                    neutral: Missing values are put into the bucket with WoE closest to 0.
                    similar: Missing values are put into the bucket with WoE closest to the bucket with only missing values.
                    passthrough: Leaves missing values untouched.
                If a dict, it must be of the following format:
                    {"<column name>": <bucket_number>}
                    This bucket number is where we will put the missing values.
            remainder: How we want the non-specified columns to be transformed. It must be in ["passthrough", "drop"].
                passthrough (Default): all columns that were not specified in "variables" will be passed through.
                drop: all remaining columns that were not specified in "variables" will be dropped.
        """  # noqa
        self.missing_treatment = missing_treatment
        self.variables = variables
        self.n_bins = n_bins
        self.specials = specials
        self.remainder = remainder
        self.get_statistics = get_statistics

    @property
    def variables_type(self):
        """
        Signals variables type supported by this bucketer.
        """
        return "numerical"

    def _get_feature_splits(self, feature, X, y, X_unfiltered=None):
        """
        Finds the splits for a single feature.

        X and y have already been preprocessed, and have specials removed.

        Args:
            feature (str): Name of the feature.
            X (pd.Series): df with single column of feature to bucket
            y (np.ndarray): array with target
            X_unfiltered (pd.Series): df with single column of feature to bucket before any filtering was applied

        Returns:
            splits, right (tuple): The splits (dict or array), and whether right=True or False.
        """
        _, boundaries = np.histogram(X.values, bins=self.n_bins)

        # np.histogram returns the min & max values of the fits
        # On transform, we use np.digitize, which means new data that is outside of this range
        # will be assigned to their own buckets.
        # To solve, we simply remove the min and max boundaries
        boundaries = boundaries[1:-1]

        if isinstance(boundaries, np.ndarray):
            boundaries = boundaries.tolist()

        return (boundaries, True)


class AgglomerativeClusteringBucketer(BaseBucketer):
    """
    The `AgglomerativeClusteringBucketer` transformer creates buckets using [sklearn.AgglomerativeClustering](https://scikit-learn.org/stable/modules/generated/sklearn.cluster.AgglomerativeClustering.html).

    Support ![badge](https://img.shields.io/badge/numerical-true-green) ![badge](https://img.shields.io/badge/categorical-false-red) ![badge](https://img.shields.io/badge/supervised-false-red)

    Example:

    ```python
    from skorecard import datasets
    from skorecard.bucketers import AgglomerativeClusteringBucketer

    specials = {"LIMIT_BAL": {"=50000": [50000], "in [20001,30000]": [20000, 30000]}}

    X, y = datasets.load_uci_credit_card(return_X_y=True)
    bucketer = AgglomerativeClusteringBucketer(n_bins = 10, variables=['LIMIT_BAL'], specials=specials)
    bucketer.fit_transform(X)
    bucketer.fit_transform(X)['LIMIT_BAL'].value_counts()
    ```
    """  # noqa

    def __init__(
        self,
        n_bins=5,
        variables=[],
        specials={},
        missing_treatment="separate",
        remainder="passthrough",
        get_statistics=True,
        **kwargs,
    ):
        """Init the class.

        Args:
            n_bins (int): Number of bins to create.
            variables (list): The features to bucket. Uses all features if not defined.
            specials: (dict) of special values that require their own binning.
                The dictionary has the following format:
                 {"<column name>" : {"name of special bucket" : <list with 1 or more values>}}
                For every feature that needs a special value, a dictionary must be passed as value.
                This dictionary contains a name of a bucket (key) and an array of unique values that should be put
                in that bucket.
                When special values are defined, they are not considered in the fitting procedure.
            missing_treatment: Defines how we treat the missing values present in the data.
                If a string, it must be one of the following options:
                    separate: Missing values get put in a separate 'Other' bucket: `-1`
                    most_risky: Missing values are put into the bucket containing the largest percentage of Class 1.
                    least_risky: Missing values are put into the bucket containing the largest percentage of Class 0.
                    most_frequent: Missing values are put into the most common bucket.
                    neutral: Missing values are put into the bucket with WoE closest to 0.
                    similar: Missing values are put into the bucket with WoE closest to the bucket with only missing values.
                    passthrough: Leaves missing values untouched.
                If a dict, it must be of the following format:
                    {"<column name>": <bucket_number>}
                    This bucket number is where we will put the missing values.
            remainder: How we want the non-specified columns to be transformed. It must be in ["passthrough", "drop"].
                passthrough (Default): all columns that were not specified in "variables" will be passed through.
                drop: all remaining columns that were not specified in "variables" will be dropped.
            kwargs: Other parameters passed to AgglomerativeBucketer
        """  # noqa
        self.variables = variables
        self.n_bins = n_bins
        self.specials = specials
        self.missing_treatment = missing_treatment
        self.remainder = remainder
        self.get_statistics = get_statistics
        self.kwargs = kwargs

    @property
    def variables_type(self):
        """
        Signals variables type supported by this bucketer.
        """
        return "numerical"

    def _get_feature_splits(self, feature, X, y, X_unfiltered=None):
        """
        Finds the splits for a single feature.

        X and y have already been preprocessed, and have specials removed.

        Args:
            feature (str): Name of the feature.
            X (pd.Series): df with single column of feature to bucket
            y (np.ndarray): array with target
            X_unfiltered (pd.Series): df with single column of feature to bucket before any filtering was applied

        Returns:
            splits, right (tuple): The splits (dict or array), and whether right=True or False.
        """
        # Fit the estimator
        ab = AgglomerativeClustering(n_clusters=self.n_bins, **self.kwargs)
        ab.fit(X.values.reshape(-1, 1), y=None)

        # Find the boundaries
        df = pd.DataFrame({"x": X.values, "label": ab.labels_}).sort_values(by="x")
        cluster_minimum_values = df.groupby("label")["x"].min().sort_values().tolist()
        cluster_maximum_values = df.groupby("label")["x"].max().sort_values().tolist()
        # take the mean of the upper boundary of a cluster and the lower boundary of the next cluster
        boundaries = [
            # Assures numbers are float and not np.float - necessary for serialization
            float(np.mean([cluster_minimum_values[i + 1], cluster_maximum_values[i]]))
            for i in range(len(cluster_minimum_values) - 1)
        ]

        if isinstance(boundaries, np.ndarray):
            boundaries = boundaries.tolist()

        return (boundaries, True)


class EqualFrequencyBucketer(BaseBucketer):
    """
    The `EqualFrequencyBucketer` transformer creates buckets with equal number of elements.

    Support: ![badge](https://img.shields.io/badge/numerical-true-green) ![badge](https://img.shields.io/badge/categorical-false-red) ![badge](https://img.shields.io/badge/supervised-false-red)

    Example:

    ```python
    from skorecard import datasets
    from skorecard.bucketers import EqualFrequencyBucketer

    X, y = datasets.load_uci_credit_card(return_X_y=True)
    bucketer = EqualFrequencyBucketer(n_bins = 10, variables=['LIMIT_BAL'])
    bucketer.fit_transform(X)
    bucketer.fit_transform(X)['LIMIT_BAL'].value_counts()
    ```
    """  # noqa

    def __init__(
        self,
        n_bins=5,
        variables=[],
        specials={},
        missing_treatment="separate",
        remainder="passthrough",
        get_statistics=True,
    ):
        """Init the class.

        Args:
            n_bins (int): Number of bins to create.
            variables (list): The features to bucket. Uses all features if not defined.
            specials: (nested) dictionary of special values that require their own binning.
                The dictionary has the following format:
                 {"<column name>" : {"name of special bucket" : <list with 1 or more values>}}
                For every feature that needs a special value, a dictionary must be passed as value.
                This dictionary contains a name of a bucket (key) and an array of unique values that should be put
                in that bucket.
                When special values are defined, they are not considered in the fitting procedure.
            missing_treatment: Defines how we treat the missing values present in the data.
                If a string, it must be one of the following options:
                    separate: Missing values get put in a separate 'Other' bucket: `-1`
                    most_risky: Missing values are put into the bucket containing the largest percentage of Class 1.
                    least_risky: Missing values are put into the bucket containing the largest percentage of Class 0.
                    most_frequent: Missing values are put into the most common bucket.
                    neutral: Missing values are put into the bucket with WoE closest to 0.
                    similar: Missing values are put into the bucket with WoE closest to the bucket with only missing values.
                    passthrough: Leaves missing values untouched.
                If a dict, it must be of the following format:
                    {"<column name>": <bucket_number>}
                    This bucket number is where we will put the missing values..
            remainder: How we want the non-specified columns to be transformed. It must be in ["passthrough", "drop"].
                passthrough (Default): all columns that were not specified in "variables" will be passed through.
                drop: all remaining columns that were not specified in "variables" will be dropped.
        """  # noqa
        self.variables = variables
        self.n_bins = n_bins
        self.specials = specials
        self.missing_treatment = missing_treatment
        self.remainder = remainder
        self.get_statistics = get_statistics

    @property
    def variables_type(self):
        """
        Signals variables type supported by this bucketer.
        """
        return "numerical"

    def _get_feature_splits(self, feature, X, y, X_unfiltered=None):
        """
        Finds the splits for a single feature.

        X and y have already been preprocessed, and have specials removed.

        Args:
            feature (str): Name of the feature.
            X (pd.Series): df with single column of feature to bucket
            y (np.ndarray): array with target
            X_unfiltered (pd.Series): df with single column of feature to bucket before any filtering was applied

        Returns:
            splits, right (tuple): The splits (dict or array), and whether right=True or False.
        """
        # Fit the estimator
        # Uses pd.qcut()
        # https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.qcut.html
        try:
            _, boundaries = pd.qcut(X, q=self.n_bins, retbins=True, duplicates="raise")
        except ValueError:
            # If there are too many duplicate values (assume a lot of filled missings)
            # this crashes - the exception drops them.
            # This means that it will return approximate quantile bins
            _, boundaries = pd.qcut(X, q=self.n_bins, retbins=True, duplicates="drop")
            warnings.warn(ApproximationWarning("Approximated quantiles - too many unique values"))

        # pd.qcut returns the min & max values of the fits
        # On transform, we use np.digitize, which means new data that is outside of this range
        # will be assigned to their own buckets.
        # To solve, we simply remove the min and max boundaries
        boundaries = boundaries[1:-1]

        if isinstance(boundaries, np.ndarray):
            boundaries = boundaries.tolist()

        # pd.qcut returns bins including right edge: (edge, edge]
        return (boundaries, True)


class DecisionTreeBucketer(BaseBucketer):
    """
    The `DecisionTreeBucketer` transformer creates buckets by training a decision tree.

    Support: ![badge](https://img.shields.io/badge/numerical-true-green) ![badge](https://img.shields.io/badge/categorical-false-red) ![badge](https://img.shields.io/badge/supervised-true-green)

    It uses [sklearn.tree.DecisionTreeClassifier](https://scikit-learn.org/stable/modules/generated/sklearn.tree.DecisionTreeClassifier.html)
    to find the splits.

    Example:

    ```python
    from skorecard import datasets
    from skorecard.bucketers import DecisionTreeBucketer
    X, y = datasets.load_uci_credit_card(return_X_y=True)

    # make sure that those cases
    specials = {
        "LIMIT_BAL":{
            "=50000":[50000],
            "in [20001,30000]":[20000,30000],
            }
    }

    dt_bucketer = DecisionTreeBucketer(variables=['LIMIT_BAL'], specials = specials)
    dt_bucketer.fit(X, y)

    dt_bucketer.fit_transform(X, y)['LIMIT_BAL'].value_counts()
    ```
    """  # noqa

    def __init__(
        self,
        variables=[],
        specials={},
        max_n_bins=100,
        missing_treatment="separate",
        min_bin_size=0.05,
        random_state=None,
        remainder="passthrough",
        get_statistics=True,
        dt_kwargs={},
    ) -> None:
        """Init the class.

        Args:
            variables (list): The features to bucket. Uses all features if not defined.
            specials (dict):  dictionary of special values that require their own binning.
                The dictionary has the following format:
                 {"<column name>" : {"name of special bucket" : <list with 1 or more values>}}
                For every feature that needs a special value, a dictionary must be passed as value.
                This dictionary contains a name of a bucket (key) and an array of unique values that should be put
                in that bucket.
                When special values are defined, they are not considered in the fitting procedure.
            min_bin_size (float): Minimum fraction of observations in a bucket. Passed directly to min_samples_leaf.
            max_n_bins (int): Maximum numbers of after the bucketing. Passed directly to max_leaf_nodes of the
                DecisionTreeClassifier.
                If specials are defined, max_leaf_nodes will be redefined to max_n_bins - (number of special bins).
                The DecisionTreeClassifier requires max_leaf_nodes>=2:
                therefore, max_n_bins  must always be >= (number of special bins + 2) if specials are defined,
                otherwise must be >=2.
            missing_treatment (str or dict): Defines how we treat the missing values present in the data.
                If a string, it must be one of the following options:
                    separate: Missing values get put in a separate 'Other' bucket: `-1`
                    most_risky: Missing values are put into the bucket containing the largest percentage of Class 1.
                    least_risky: Missing values are put into the bucket containing the largest percentage of Class 0.
                    most_frequent: Missing values are put into the most common bucket.
                    neutral: Missing values are put into the bucket with WoE closest to 0.
                    similar: Missing values are put into the bucket with WoE closest to the bucket with only missing values.
                    passthrough: Leaves missing values untouched.
                If a dict, it must be of the following format:
                    {"<column name>": <bucket_number>}
                    This bucket number is where we will put the missing values.
            random_state (int): The random state, Passed directly to DecisionTreeClassifier
            remainder (str): How we want the non-specified columns to be transformed. It must be in ["passthrough", "drop"].
                passthrough (Default): all columns that were not specified in "variables" will be passed through.
                drop: all remaining columns that were not specified in "variables" will be dropped.
            dt_kwargs: Other parameters passed to DecisionTreeClassifier
        """  # noqa
        self.variables = variables
        self.specials = specials
        self.dt_kwargs = dt_kwargs
        self.max_n_bins = max_n_bins
        self.missing_treatment = missing_treatment
        self.min_bin_size = min_bin_size
        self.random_state = random_state
        self.remainder = remainder
        self.get_statistics = get_statistics
        self.dt_kwargs.update({"random_state": self.random_state})

        check_args(dt_kwargs, DecisionTreeClassifier)

    @property
    def variables_type(self):
        """
        Signals variables type supported by this bucketer.
        """
        return "numerical"

    def _get_feature_splits(self, feature, X, y, X_unfiltered=None):
        """
        Finds the splits for a single feature.

        X and y have already been preprocessed, and have specials removed.

        Args:
            feature (str): Name of the feature.
            X (pd.Series): df with single column of feature to bucket
            y (np.ndarray): array with target
            X_unfiltered (pd.Series): df with single column of feature to bucket before any filtering was applied

        Returns:
            splits, right (tuple): The splits (dict or array), and whether right=True or False.
        """
        # Make sure max_n_bins settings is correct
        n_special_bins = 0
        if feature in self.specials.keys():
            n_special_bins = len(self.specials[feature])
            if (self.max_n_bins - n_special_bins) <= 1:
                raise ValueError(
                    f"max_n_bins must be at least = the number of special bins + 2: set a value "
                    f"max_n_bins>= {n_special_bins+2} (currently max_n_bins={self.max_n_bins})"
                )

        # If the data contains only specials,
        # Then don't use any splits
        if X.shape[0] == 0:
            splits = []
        else:
            # If the specials are excluded, make sure that the bin size is rescaled.
            frac_left = X.shape[0] / X_unfiltered.shape[0]
            min_bin_size = self.min_bin_size / frac_left

            if min_bin_size > 0.5:
                min_bin_size = 0.5

            binner = DecisionTreeClassifier(
                max_leaf_nodes=(self.max_n_bins - n_special_bins),
                min_samples_leaf=min_bin_size,
                **self.dt_kwargs,
            )
            binner.fit(X.values.reshape(-1, 1), y)

            # Extract fitted boundaries
            splits = np.unique(binner.tree_.threshold[binner.tree_.feature != _tree.TREE_UNDEFINED])

        # Note for trees we use right=False
        return (splits, False)


class OrdinalCategoricalBucketer(BaseBucketer):
    """
    The `OrdinalCategoricalBucketer` replaces categories by ordinal numbers.

    Support ![badge](https://img.shields.io/badge/numerical-false-red) ![badge](https://img.shields.io/badge/categorical-true-green) ![badge](https://img.shields.io/badge/supervised-true-green)

    When `sort_by_target` is `false` the buckets are assigned in order of frequency.
    When `sort_by_target` is `true` the buckets are ordered based on the mean of the target per category.

    For example, if for a variable `colour` the means of the target
    for `blue`, `red` and `grey` is `0.5`, `0.8` and `0.1` respectively,
    `grey` will be the first bucket (`0`), blue the second (`1`) and
    `red` the third (`3`). If new data contains unknown labels (f.e. yellow),
    they will be replaced by the 'Other' bucket (`-2`),
    and if new data contains missing values, they will be replaced by the 'Missing' bucket (`-1`).

    Example:

    ```python
    from skorecard import datasets
    from skorecard.bucketers import OrdinalCategoricalBucketer

    X, y = datasets.load_uci_credit_card(return_X_y=True)
    bucketer = OrdinalCategoricalBucketer(variables=['EDUCATION'])
    bucketer.fit_transform(X, y)
    bucketer = OrdinalCategoricalBucketer(max_n_categories=2, variables=['EDUCATION'])
    bucketer.fit_transform(X, y)
    ```

    Credits: Code & ideas adapted from:

    - feature_engine.categorical_encoders.OrdinalCategoricalEncoder
    - feature_engine.categorical_encoders.RareLabelCategoricalEncoder

    """  # noqa

    def __init__(
        self,
        tol=0.05,
        max_n_categories=None,
        variables=[],
        specials={},
        encoding_method="frequency",
        missing_treatment="separate",
        remainder="passthrough",
        get_statistics=True,
    ):
        """
        Init the class.

        Args:
            tol (float): the minimum frequency a label should have to be considered frequent.
                Categories with frequencies lower than tol will be grouped together (in the 'other' bucket).
            max_n_categories (int): the maximum number of categories that should be considered frequent.
                If None, all categories with frequency above the tolerance (tol) will be
                considered.
            variables (list): The features to bucket. Uses all features if not defined.
            specials (dict): (nested) dictionary of special values that require their own binning.
                The dictionary has the following format:
                 {"<column name>" : {"name of special bucket" : <list with 1 or more values>}}
                For every feature that needs a special value, a dictionary must be passed as value.
                This dictionary contains a name of a bucket (key) and an array of unique values that should be put
                in that bucket.
                When special values are defined, they are not considered in the fitting procedure.
            encoding_method (string): encoding method.
                - "frequency" (default): orders the buckets based on the frequency of observations in the bucket.
                    The lower the number of the bucket the most frequent are the observations in that bucket.
                - "ordered": orders the buckets based on the average class 1 rate in the bucket.
                    The lower the number of the bucket the lower the fraction of class 1 in that bucket.
            missing_treatment (str or dict): Defines how we treat the missing values present in the data.
                If a string, it must be one of the following options:
                    separate: Missing values get put in a separate 'Other' bucket: `-1`
                    most_risky: Missing values are put into the bucket containing the largest percentage of Class 1.
                    least_risky: Missing values are put into the bucket containing the largest percentage of Class 0.
                    most_frequent: Missing values are put into the most common bucket.
                    neutral: Missing values are put into the bucket with WoE closest to 0.
                    similar: Missing values are put into the bucket with WoE closest to the bucket with only missing values.
                    passthrough: Leaves missing values untouched.
                If a dict, it must be of the following format:
                    {"<column name>": <bucket_number>}
                    This bucket number is where we will put the missing values.
            remainder (str): How we want the non-specified columns to be transformed. It must be in ["passthrough", "drop"].
                passthrough (Default): all columns that were not specified in "variables" will be passed through.
                drop: all remaining columns that were not specified in "variables" will be dropped.
        """  # noqa
        self.tol = tol
        self.max_n_categories = max_n_categories
        self.variables = variables
        self.specials = specials
        self.encoding_method = encoding_method
        self.missing_treatment = missing_treatment
        self.remainder = remainder
        self.get_statistics = get_statistics

    @property
    def variables_type(self):
        """
        Signals variables type supported by this bucketer.
        """
        return "categorical"

    def _get_feature_splits(self, feature, X, y, X_unfiltered=None):
        """
        Finds the splits for a single feature.

        X and y have already been preprocessed, and have specials removed.

        Args:
            feature (str): Name of the feature.
            X (pd.Series): df with single column of feature to bucket
            y (np.ndarray): array with target
            X_unfiltered (pd.Series): df with single column of feature to bucket before any filtering was applied

        Returns:
            splits, right (tuple): The splits (dict or array), and whether right=True or False.
        """
        normalized_counts = None

        if y is None:
            y = pd.Series(None)
        elif not (isinstance(y, pd.Series) or isinstance(y, pd.DataFrame)):
            y = pd.Series(y)
        else:
            raise AssertionError("something wrong with format of y")

        X_y = pd.concat([X, y], axis=1)
        X_y.columns = [feature, "target"]

        if self.encoding_method == "ordered":
            if y is None:
                raise ValueError("To use encoding_method=='ordered', y cannot be None.")

            normalized_counts = X_y[feature].value_counts(normalize=True)
            cats = X_y.groupby([feature])["target"].mean().sort_values(ascending=True).index
            normalized_counts = normalized_counts[cats]

        elif self.encoding_method == "frequency":
            normalized_counts = X_y[feature].value_counts(normalize=True)

        # Limit number of categories if set.
        normalized_counts = normalized_counts[: self.max_n_categories]
        # Remove less frequent categories
        normalized_counts = normalized_counts[normalized_counts >= self.tol]

        # Determine Ordinal Encoder based on ordered labels
        # Note we start at 1, to be able to encode missings as 0.
        mapping = dict(zip(normalized_counts.index, range(0, len(normalized_counts))))

        # Note that right is set to True, but this is not used at all for categoricals
        return (mapping, True)


class AsIsCategoricalBucketer(BaseBucketer):
    """
    The `AsIsCategoricalBucketer` treats unique values as categories.

    Support: ![badge](https://img.shields.io/badge/numerical-false-red) ![badge](https://img.shields.io/badge/categorical-true-green) ![badge](https://img.shields.io/badge/supervised-false-blue)

    It will assign each a bucket number in the order of appearance.
    If new data contains new, unknown labels they will be replaced by 'Other'.

    This is bucketer is useful when you have data that is already sufficiented bucketed,
    but you would like to be able to bucket new data in the same way.

    Example:

    ```python
    from skorecard import datasets
    from skorecard.bucketers import AsIsCategoricalBucketer

    X, y = datasets.load_uci_credit_card(return_X_y=True)
    bucketer = AsIsCategoricalBucketer(variables=['EDUCATION'])
    bucketer.fit_transform(X)
    ```
    """  # noqa

    def __init__(
        self, variables=[], specials={}, missing_treatment="separate", remainder="passthrough", get_statistics=True
    ):
        """Init the class.

        Args:
            variables (list): The features to bucket. Uses all features if not defined.
            specials: (nested) dictionary of special values that require their own binning.
                The dictionary has the following format:
                 {"<column name>" : {"name of special bucket" : <list with 1 or more values>}}
                For every feature that needs a special value, a dictionary must be passed as value.
                This dictionary contains a name of a bucket (key) and an array of unique values that should be put
                in that bucket.
                When special values are defined, they are not considered in the fitting procedure.
            missing_treatment: Defines how we treat the missing values present in the data.
                If a string, it must be one of the following options:
                    separate: Missing values get put in a separate 'Other' bucket: `-1`
                    most_risky: Missing values are put into the bucket containing the largest percentage of Class 1.
                    least_risky: Missing values are put into the bucket containing the largest percentage of Class 0.
                    most_frequent: Missing values are put into the most common bucket.
                    neutral: Missing values are put into the bucket with WoE closest to 0.
                    similar: Missing values are put into the bucket with WoE closest to the bucket with only missing values.
                    passthrough: Leaves missing values untouched.
                If a dict, it must be of the following format:
                    {"<column name>": <bucket_number>}
                    This bucket number is where we will put the missing values.
            remainder: How we want the non-specified columns to be transformed. It must be in ["passthrough", "drop"].
                passthrough (Default): all columns that were not specified in "variables" will be passed through.
                drop: all remaining columns that were not specified in "variables" will be dropped.
        """  # noqa
        self.variables = variables
        self.specials = specials
        self.missing_treatment = missing_treatment
        self.remainder = remainder
        self.get_statistics = get_statistics

    @property
    def variables_type(self):
        """
        Signals variables type supported by this bucketer.
        """
        return "categorical"

    def _get_feature_splits(self, feature, X, y, X_unfiltered=None):
        """
        Finds the splits for a single feature.

        X and y have already been preprocessed, and have specials removed.

        Args:
            feature (str): Name of the feature.
            X (pd.Series): df with single column of feature to bucket
            y (np.ndarray): array with target
            X_unfiltered (pd.Series): df with single column of feature to bucket before any filtering was applied

        Returns:
            splits, right (tuple): The splits (dict or array), and whether right=True or False.
        """
        unq = X.unique().tolist()
        mapping = dict(zip(unq, range(0, len(unq))))

        # Note that right is set to True, but this is not used at all for categoricals
        return (mapping, True)


class AsIsNumericalBucketer(BaseBucketer):
    """
    The `AsIsNumericalBucketer` transformer creates buckets by treating the existing unique values as boundaries.

    Support: ![badge](https://img.shields.io/badge/numerical-true-green) ![badge](https://img.shields.io/badge/categorical-false-red) ![badge](https://img.shields.io/badge/supervised-false-blue)

    This is bucketer is useful when you have data that is already sufficiented bucketed,
    but you would like to be able to bucket new data in the same way.

    Example:

    ```python
    from skorecard import datasets
    from skorecard.bucketers import AsIsNumericalBucketer

    X, y = datasets.load_uci_credit_card(return_X_y=True)
    bucketer = AsIsNumericalBucketer(variables=['LIMIT_BAL'])
    bucketer.fit_transform(X)
    ```
    """  # noqa

    def __init__(
        self,
        right=True,
        variables=[],
        specials={},
        missing_treatment="separate",
        remainder="passthrough",
        get_statistics=True,
    ):
        """
        Init the class.

        Args:
            right (boolean): Is the right value included in a range (default) or is 'up to not but including'.
                For example, if you have [5, 10], the ranges for right=True would be (-Inf, 5], (5, 10], (10, Inf]
                or [-Inf, 5), [5, 10), [10, Inf) for right=False
            variables (list): The features to bucket. Uses all features if not defined.
            specials (dict): (nested) dictionary of special values that require their own binning.
                The dictionary has the following format:
                 {"<column name>" : {"name of special bucket" : <list with 1 or more values>}}
                For every feature that needs a special value, a dictionary must be passed as value.
                This dictionary contains a name of a bucket (key) and an array of unique values that should be put
                in that bucket.
                When special values are defined, they are not considered in the fitting procedure.
            missing_treatment (str or dict): Defines how we treat the missing values present in the data.
                If a string, it must be one of the following options:
                    separate: Missing values get put in a separate 'Other' bucket: `-1`
                    most_risky: Missing values are put into the bucket containing the largest percentage of Class 1.
                    least_risky: Missing values are put into the bucket containing the largest percentage of Class 0.
                    most_frequent: Missing values are put into the most common bucket.
                    neutral: Missing values are put into the bucket with WoE closest to 0.
                    similar: Missing values are put into the bucket with WoE closest to the bucket with only missing values.
                    passthrough: Leaves missing values untouched.
                If a dict, it must be of the following format:
                    {"<column name>": <bucket_number>}
                    This bucket number is where we will put the missing values..
            remainder (str): How we want the non-specified columns to be transformed. It must be in ["passthrough", "drop"].
                passthrough (Default): all columns that were not specified in "variables" will be passed through.
                drop: all remaining columns that were not specified in "variables" will be dropped.
        """  # noqa
        self.right = right
        self.variables = variables
        self.specials = specials
        self.missing_treatment = missing_treatment
        self.remainder = remainder
        self.get_statistics = get_statistics

    @property
    def variables_type(self):
        """
        Signals variables type supported by this bucketer.
        """
        return "numerical"

    def _get_feature_splits(self, feature, X, y, X_unfiltered=None):
        """
        Finds the splits for a single feature.

        X and y have already been preprocessed, and have specials removed.

        Args:
            feature (str): Name of the feature.
            X (pd.Series): df with single column of feature to bucket
            y (np.ndarray): array with target
            X_unfiltered (pd.Series): df with single column of feature to bucket before any filtering was applied

        Returns:
            splits, right (tuple): The splits (dict or array), and whether right=True or False.
        """
        boundaries = X.unique().tolist()
        boundaries.sort()

        if len(boundaries) > 100:
            msg = f"The column '{feature}' has more than 100 unique values "
            msg += "and cannot be used with the AsIsBucketer."
            msg += "Apply a different bucketer first."
            raise NotPreBucketedError(msg)

        return (boundaries, self.right)


class UserInputBucketer(BaseBucketer):
    """
    The `UserInputBucketer` transformer creates buckets by implementing user-defined boundaries.

    Support: ![badge](https://img.shields.io/badge/numerical-true-green) ![badge](https://img.shields.io/badge/categorical-true-green) ![badge](https://img.shields.io/badge/supervised-false-blue)

    This is a special bucketer that is not fitted but rather relies
    on pre-defined user input. The most common use-case is loading
    bucket mapping information previously fitted by other bucketers.

    Example:

    ```python
    from skorecard import datasets
    from skorecard.bucketers import AgglomerativeClusteringBucketer, UserInputBucketer

    X, y = datasets.load_uci_credit_card(return_X_y=True)

    ac_bucketer = AgglomerativeClusteringBucketer(n_bins=3, variables=['LIMIT_BAL'])
    ac_bucketer.fit(X)
    mapping = ac_bucketer.features_bucket_mapping_

    ui_bucketer = UserInputBucketer(mapping)
    new_X = ui_bucketer.fit_transform(X)
    assert len(new_X['LIMIT_BAL'].unique()) == 3

    #Map some values to the special buckets
    specials = {
        "LIMIT_BAL":{
            "=50000":[50000],
            "in [20001,30000]":[20000,30000],
            }
    }

    ac_bucketer = AgglomerativeClusteringBucketer(n_bins=3, variables=['LIMIT_BAL'], specials = specials)
    ac_bucketer.fit(X)
    mapping = ac_bucketer.features_bucket_mapping_

    ui_bucketer = UserInputBucketer(mapping)
    new_X = ui_bucketer.fit_transform(X)
    assert len(new_X['LIMIT_BAL'].unique()) == 5
    ```

    """  # noqa

    def __init__(
        self, features_bucket_mapping=None, variables: List = [], remainder="passthrough", get_statistics=True
    ) -> None:
        """
        Initialise the user-defined boundaries with a dictionary.

        Notes:
        - features_bucket_mapping is stored without the trailing underscore (_) because it is not fitted.

        Args:
            features_bucket_mapping (None, Dict, FeaturesBucketMapping, str or Path): Contains the feature name and boundaries
                defined for this feature.
                If a dict, it will be converted to an internal FeaturesBucketMapping object.
                If a string or path, which will attempt to load the file as a yaml and convert to FeaturesBucketMapping object.
            variables (list): The features to bucket. Uses all features in features_bucket_mapping if not defined.
            remainder (str): How we want the non-specified columns to be transformed. It must be in ["passthrough", "drop"].
                passthrough (Default): all columns that were not specified in "variables" will be passed through.
                drop: all remaining columns that were not specified in "variables" will be dropped.
        """  # noqa
        # Assigning the variable in the init to the attribute with the same name is a requirement of
        # sklearn.base.BaseEstimator. See the notes in
        # https://scikit-learn.org/stable/modules/generated/sklearn.base.BaseEstimator.html#sklearn.base.BaseEstimator
        self.features_bucket_mapping = features_bucket_mapping
        self.remainder = remainder
        self.get_statistics = get_statistics

        self.variables = variables

        if features_bucket_mapping is None:
            self.features_bucket_mapping_ = FeaturesBucketMapping()
        elif isinstance(features_bucket_mapping, str):
            buckets_yaml = yaml.safe_load(open(features_bucket_mapping))
            self.features_bucket_mapping_ = FeaturesBucketMapping(buckets_yaml)
        elif isinstance(features_bucket_mapping, dict):
            self.features_bucket_mapping_ = FeaturesBucketMapping(features_bucket_mapping)
        elif isinstance(features_bucket_mapping, FeaturesBucketMapping):
            self.features_bucket_mapping_ = features_bucket_mapping
        else:
            try:
                buckets_yaml = yaml.safe_load(features_bucket_mapping)
                self.features_bucket_mapping_ = FeaturesBucketMapping(buckets_yaml)
            except Exception:
                raise TypeError(
                    "'features_bucket_mapping' must be a None, dict, str, path, or FeaturesBucketMapping instance"
                )

    def fit(self, X, y=None):
        """Init the class."""
        X = ensure_dataframe(X)
        if y is not None:
            assert len(y) == X.shape[0], "y and X not same length"
            # Store the classes seen during fit
            self.classes_ = unique_labels(y)

        # scikit-learn requires checking that X has same shape on transform
        # this is because scikit-learn is still positional based (no column names used)
        self.n_train_features_ = X.shape[1]

        # bucket tables can only be computed on fit().
        # so a user will have to .fit() if she/he wants .plot_buckets() and .bucket_table()
        self.bucket_tables_ = {}

        # and if user did not specify any variables
        # use all the variables defined in the features_bucket_mapping
        if self.variables == []:
            self.variables_ = list(self.features_bucket_mapping_.maps.keys())

        for feature in self.variables_:
            # Calculate the bucket table
            self.bucket_tables_[feature] = build_bucket_table(
                X,
                y,
                column=feature,
                bucket_mapping=self.features_bucket_mapping_.get(feature),
            )

        self._generate_summary(X, y)

        return self

    def _more_tags(self):
        """
        Estimator tags are annotations of estimators that allow programmatic inspection of their capabilities.

        See https://scikit-learn.org/stable/developers/develop.html#estimator-tags
        """  # noqa
        return {"binary_only": True, "allow_nan": True, "requires_fit": False}



import numpy as np

from skorecard.bucketers import AsIsCategoricalBucketer


def test_correct_output(df):
    """Test that correct use of CatBucketTransformer returns expected results."""
    X = df
    y = df["default"].values

    asb = AsIsCategoricalBucketer(variables=["EDUCATION"])
    asb.fit(X, y)
    X_trans = asb.transform(X)
    assert len(X["EDUCATION"].unique()) == len(X_trans["EDUCATION"].unique())


def test_specials(df):
    """Test specials get assigned to the right bin."""
    X = df[["EDUCATION"]]
    y = df["default"]

    asb = AsIsCategoricalBucketer(variables=["EDUCATION"], specials={"EDUCATION": {"ed 0": [1]}})
    asb.fit(X, y)

    X_transform = asb.transform(X)
    # Make sure value 1 is assigned special bucket
    assert np.unique(X_transform[X["EDUCATION"] == 1].values)[0] == -3


def test_missing_default(df_with_missings) -> None:
    """Test that missing values are assigned to the right bucket."""
    X = df_with_missings
    y = df_with_missings["default"].values

    bucketer = AsIsCategoricalBucketer(variables=["EDUCATION"])
    X["EDUCATION_trans"] = bucketer.fit_transform(X[["EDUCATION"]], y)

    assert len(X["EDUCATION_trans"].unique()) == 8  # 7 unique values + 1 for NAs



import numpy as np
import pandas as pd
import pytest

from skorecard import datasets
from skorecard.bucketers import DecisionTreeBucketer


@pytest.fixture()
def df():
    """Generate dataframe."""
    return datasets.load_uci_credit_card(as_frame=True)


def test_transform(df):
    """Test that the correct shape is returned."""
    X = df[["LIMIT_BAL", "BILL_AMT1"]]
    y = df["default"]

    tbt = DecisionTreeBucketer(
        variables=["LIMIT_BAL", "BILL_AMT1"],
        dt_kwargs={"criterion": "entropy", "min_impurity_decrease": 0.001},
    )
    tbt.fit(X, y)

    assert tbt.transform(X).shape == X.shape


def test_running_twice(df):
    """Test that running same bucketer twice does not change the bins.

    Context is that OptimalBucketer wraps optbinning.OptimalBinning, which also does prebinning using a DecisionTree.
    It's a lot more work to extract from optbinning only the optimization algorithm, so we choose to
    feed to OptimalBucketer already pre-binned buckets.

    We need to make sure that running it twice doesn't change the buckets.
    """
    X = df[["LIMIT_BAL", "BILL_AMT1"]]
    y = df["default"]

    tbt = DecisionTreeBucketer(variables=["LIMIT_BAL", "BILL_AMT1"])
    tbt.fit(X, y)
    X_prebucketed = tbt.transform(X)

    tbt2 = DecisionTreeBucketer(variables=["LIMIT_BAL", "BILL_AMT1"])
    tbt2.fit(X_prebucketed, y)
    X_bucketed = tbt2.transform(X_prebucketed)

    assert X_prebucketed["BILL_AMT1"].value_counts().equals(X_bucketed["BILL_AMT1"].value_counts())

    # But what about using a different random_state?
    tbt3 = DecisionTreeBucketer(variables=["LIMIT_BAL", "BILL_AMT1"], random_state=1)
    tbt3.fit(X_prebucketed, y)
    X_bucketed = tbt3.transform(X_prebucketed)

    assert X_prebucketed["BILL_AMT1"].value_counts().equals(X_bucketed["BILL_AMT1"].value_counts())


def test_specials_filters(df):
    """Test that when adding specials,the binner performs as expected.

    Context: special values should be binned in their own bin.
    """
    X = df[["LIMIT_BAL", "BILL_AMT1"]]
    y = df["default"]

    specials = {"LIMIT_BAL": {"=50000": [50000], "in [20001,30000]": [20000, 30000]}}

    f = DecisionTreeBucketer._filter_specials_for_fit

    X_flt, y_flt = f(X["LIMIT_BAL"], y, specials=specials["LIMIT_BAL"])

    assert X_flt[X_flt.isin([20000, 30000, 50000])].shape[0] == 0
    assert y_flt.equals(y[~X["LIMIT_BAL"].isin([20000, 30000, 50000])])


def test_all_data_is_special(df):
    """Test that when all the data is in the special buckets, the code does not crash."""
    X = df[["LIMIT_BAL", "BILL_AMT1"]]
    y = df["default"]

    X["all_specials"] = pd.Series([999 for i in range(X.shape[0])])
    X["all_nans"] = pd.Series([np.nan for i in range(X.shape[0])])
    specials = {"all_specials": {"=999": [999]}}

    tbt = DecisionTreeBucketer(variables=["all_specials", "all_nans"], specials=specials)
    tbt.fit(X, y)
    X_prebucketed = tbt.transform(X)

    assert X_prebucketed["all_nans"].unique() == -1
    assert X_prebucketed["all_specials"].unique() == -3


def test_missing_default(df_with_missings) -> None:
    """Test that missing values are assigned to the right bucket."""
    X = df_with_missings
    y = df_with_missings["default"].values

    BUCK = DecisionTreeBucketer(variables=["LIMIT_BAL"], random_state=1)
    BUCK.fit(X, y)
    X["LIMIT_BAL_trans"] = BUCK.transform(X)["LIMIT_BAL"]

    missing_bucket = [f for f in BUCK.features_bucket_mapping_.get("LIMIT_BAL").labels.keys()][-1]
    assert BUCK.features_bucket_mapping_.get("LIMIT_BAL").labels[missing_bucket] == "Missing"
    assert X[np.isnan(X["LIMIT_BAL"])].shape[0] == X[X["LIMIT_BAL_trans"] == missing_bucket].shape[0]


def test_missing_manual(df_with_missings) -> None:
    """Test that missing values are assigned to the right bucket."""
    X = df_with_missings
    y = df_with_missings["default"].values

    bucketer = DecisionTreeBucketer(variables=["LIMIT_BAL"], random_state=1, missing_treatment={"LIMIT_BAL": 0})
    bucketer.fit(X, y)
    X["LIMIT_BAL_trans"] = bucketer.transform(X)["LIMIT_BAL"]

    assert X[np.isnan(X["LIMIT_BAL_trans"])]["LIMIT_BAL"].sum() == 0



from functools import reduce

import numpy as np
import pandas as pd
from category_encoders.woe import WOEEncoder
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.utils.validation import check_is_fitted

from skorecard import Skorecard
from skorecard.preprocessing import WoeEncoder


def calibrate_to_master_scale(y_pred, *, pdo, ref_score, ref_odds, epsilon=1e-6):
    """Calibrate the score to the master scale.

    It's common practice to represent the model predictions on a 'master scale'.
    The rescaling step is defined as follows:
        A reference score is set that corresponds to some reference odds of being class 0.
        By adding pdo (points to double the odds), the odds will double.

    Example:
        ref_score = 300 for ref_odds 50:1, and pdo = 25
        if the odds become 100:1, the score is 325.
        if the odds become 200:1, the score is 350.
        if the odds become 25:1, the score is 275.

    Args:
        y_pred: predicted probabilities
        pdo: number of points necessary to double the odds
        ref_score: reference score set for the reference odds
        ref_odds: odds that correspond to the ref_score
        epsilon: float (1e-6), correction to avoid infinite odds if y_pred = 0

    Returns:
        pd.Series (or np.array if y_pred was a np.array), rescaled integer scores
    """
    is_np = False
    if isinstance(y_pred, np.ndarray):
        y_pred = pd.Series(y_pred)
        is_np = True
    odd_func = lambda x: (1 - x + epsilon) / (x + epsilon)
    odds = y_pred.apply(odd_func)

    factor = pdo / np.log(2)
    offset = ref_score - factor * np.log(ref_odds)

    master_scale = odds.apply(lambda x: _map_to_scale(x, factor, offset))

    if is_np:
        return master_scale.values
    else:
        return master_scale


def _map_to_scale(x, factor, offset):
    """Helper function."""
    try:
        return int(round(factor * np.log(x) + offset))
    except ValueError:
        return -999


class ScoreCardPoints(BaseEstimator, TransformerMixin):
    """Transformer to map the the buckets from the skorecard model and maps them to the rescaled points.

    Examples:

    ```python
    from skorecard import Skorecard
    from skorecard.rescale import ScoreCardPoints
    from skorecard.datasets import load_uci_credit_card

    X,y = load_uci_credit_card(return_X_y=True)
    model = Skorecard(variables = ["LIMIT_BAL", "BILL_AMT1","EDUCATION", "MARRIAGE"])
    model.fit(X, y)

    scp = ScoreCardPoints(model)
    scp.transform(X)
    ```

    """

    def __init__(self, skorecard_model, *, pdo=20, ref_score=100, ref_odds=1):
        """Initialize the transformer.

        Args:
            skorecard_model: the fitted Skorecard class
            pdo: number of points necessary to double the odds
            ref_score: reference score set for the reference odds
            ref_odds: odds that correspond to the ref_score
        """
        assert isinstance(skorecard_model, Skorecard), (
            f"The skorecard_model must be an instance of "
            f"skorecard.Skorecard, got {skorecard_model.__class__.__name__} instead."
        )
        check_is_fitted(skorecard_model)
        self.skorecard_model = skorecard_model
        # self.pipeline = skorecard_model.pipeline
        self.pdo = pdo
        self.ref_score = ref_score
        self.ref_odds = ref_odds
        self._get_pipeline_elements()
        self._calculate_scorecard_points()

    def _get_pipeline_elements(self):
        bucketers = self.skorecard_model.pipeline_.named_steps["bucketer"]
        woe_enc = self.skorecard_model.pipeline_.named_steps["encoder"]
        self.features = self.skorecard_model.variables
        self.model = self.skorecard_model.pipeline_.named_steps["model"]

        assert hasattr(self.model, "predict_proba"), (
            f"Expected a model at the end of the pipeline, " f"got {self.model.__class__}"
        )
        if not (isinstance(woe_enc, WoeEncoder) or isinstance(woe_enc, WOEEncoder)):
            raise ValueError("Pipeline must have WoE encoder")

        fbm = bucketers.features_bucket_mapping_

        if len(self.features) == 0:
            # there is no feature selector
            self.features = fbm.columns
        woe_dict = woe_enc.mapping

        self.buckets = {k: fbm.get(k) for k in fbm.columns if k in self.features}
        self.woes = {k: woe_dict[k] for k in woe_dict.keys() if k in self.features}

    def _calculate_scorecard_points(self):
        # Put together the features in a list of tables, containing all the buckets.
        list_dfs = list()
        for ix, col in enumerate(self.features):
            df_ = (
                pd.concat([pd.Series(self.buckets[col].labels), pd.Series(self.woes[col])], axis=1)
                .reset_index()
                .rename(columns={"index": "bin_index", 0: "map", 1: "woe"})
            )
            df_.loc[:, "feature"] = col

            df_.loc[:, "coef"] = self.model.coef_[0][ix]
            #
            list_dfs.append(df_)

        # Reduce the list of tables, to build the final scorecard feature points
        scorecard = reduce(lambda x, y: pd.concat([x, y]), list_dfs)
        scorecard = pd.concat(
            [
                scorecard,
                pd.DataFrame.from_records(
                    [{"feature": "Intercept", "coef": self.model.intercept_[0], "bin_index": 0, "map": 0, "woe": 0}]
                ),
            ],
            ignore_index=True,
        )

        #     return buckets, woes
        scorecard["contribution"] = scorecard["woe"] * scorecard["coef"]

        self.scorecard = _scale_scorecard(
            scorecard, pdo=self.pdo, ref_score=self.ref_score, ref_odds=self.ref_odds, features=self.features
        )

        self.points_mapper = dict()
        for feat in self.scorecard["feature"].unique():
            one_feat_df = self.scorecard.loc[self.scorecard["feature"] == feat, ["bin_index", "Points"]]
            self.points_mapper[feat] = {
                k: v for k, v in zip(one_feat_df["bin_index"].values, one_feat_df["Points"].values)
            }

    def get_scorecard_points(self):
        """Get the scorecard points."""
        return self.scorecard

    def transform(self, X):
        """Transform the features to the points."""
        X_buckets = self.skorecard_model.pipeline_.named_steps["bucketer"].transform(X)

        bin_points = pd.concat(
            [
                X_buckets[feat].map(self.points_mapper[feat])
                for feat in self.points_mapper.keys()
                if feat != "Intercept"
            ],
            axis=1,
        )
        bin_points.index = X.index

        return bin_points


def _scale_scorecard(df, *, pdo, ref_score, ref_odds, features):
    """Equations to scale the feature scorecards.

    Args:
        df: Pandas DataFrame
        pdo: number of points necessary to double the odds
        ref_score: reference score set for the reference odds
        ref_odds: odds that correspond to the ref_score
        features (list): The list of features

    """
    df = df[~df["woe"].isnull()]

    factor = pdo / np.log(2)
    offset = ref_score - factor * np.log(ref_odds)

    intercept = df.loc[df["feature"] == "Intercept"]["coef"].values[0]
    n = len(features)

    df_out = df.copy()
    df_out["Points"] = (offset / n) - (df_out["contribution"] + (intercept / n)) * factor
    df_out.loc[df_out["feature"] == "Intercept", "Points"] = 0

    df_out["Points"] = df_out["Points"].apply(lambda x: int(round(x, 0)))

    return df_out



import numpy as np
import pandas as pd
import pytest

import skorecard.reporting.report
from skorecard.bucketers import DecisionTreeBucketer
from skorecard.metrics import metrics


@pytest.fixture()
def X_y():
    """Set of X,y for testing the transformers."""
    X = np.array(
        [[0, 1], [1, 0], [0, 0], [3, 2], [0, 1], [1, 2], [2, 0], [2, 1], [0, 0]],
        np.int32,
    )
    y = np.array([0, 0, 0, 1, 1, 1, 0, 0, 1])

    return X, y


@pytest.fixture()
def X1_X2():
    """Set of dataframes to test psi."""
    X1 = pd.DataFrame(
        [[0, 1], [1, 0], [0, 0], [3, 2], [0, 1], [1, 2], [2, 0], [2, 1], [0, 0]], columns=["col1", "col2"]
    )
    X2 = pd.DataFrame(
        [[0, 2], [3, 0], [0, 0], [1, 2], [0, 4], [2, 1], [1, 1], [2, 1], [1, 1]], columns=["col1", "col2"]
    )

    return X1, X2


def test_iv_on_array(X_y):
    """Test the IV calculation for two arrays."""
    X, y = X_y
    X = pd.DataFrame(X, columns=["0", "1"])

    np.testing.assert_almost_equal(metrics._IV_score(y, X["0"]), 5.307, decimal=2)

    np.testing.assert_almost_equal(metrics._IV_score(y, X["1"]), 4.635, decimal=2)


def test_psi_zero(df):
    """Test that PSI on same array is zero."""
    features = ["LIMIT_BAL", "BILL_AMT1"]
    X = df[features]
    y = df["default"]

    X_bins = DecisionTreeBucketer(variables=features).fit_transform(X, y)

    psi_vals = skorecard.reporting.report.psi(X_bins, X_bins)

    assert set(psi_vals.keys()) == set(X_bins.columns)
    assert all([val == 0 for val in psi_vals.values()])


def test_psi_values(X1_X2):
    """Assert PSi values match expectations."""
    X1, X2 = X1_X2
    expected_psi = {"col1": 0.0773, "col2": 0.965}
    psi_vals = skorecard.reporting.report.psi(X1, X2)

    np.testing.assert_array_almost_equal(pd.Series(expected_psi).values, pd.Series(psi_vals).values, decimal=2)


def test_IV_values(X_y):
    """Assert IV values match expectations."""
    X, y = X_y
    random_index = [2 * x for x in range(0, len(y))]
    X = pd.DataFrame(X, columns=["col1", "col2"], index=random_index)
    expected_iv = {"col1": 5.307, "col2": 4.635}
    iv_vals = skorecard.reporting.report.iv(X, y)
    np.testing.assert_array_almost_equal(pd.Series(expected_iv).values, pd.Series(iv_vals).values, decimal=2)

    # Make sure these are still accurate if y is a pd.Series with the same
    # non-continuous indices as X.
    y = pd.Series(y, index=random_index)
    iv_vals = skorecard.reporting.report.iv(X, y)
    np.testing.assert_array_almost_equal(pd.Series(expected_iv).values, pd.Series(iv_vals).values, decimal=2)



import collections
import itertools

import pytest
from category_encoders.woe import WOEEncoder
from sklearn.linear_model import LogisticRegression
from sklearn.model_selection import cross_val_score
from sklearn.pipeline import make_pipeline
from sklearn.preprocessing import StandardScaler
from sklearn.utils import estimator_checks

from skorecard.bucketers import DecisionTreeBucketer, OptimalBucketer
from skorecard.pipeline import BucketingProcess
from tests.conftest import CLASSIFIERS, TRANSFORMERS
import warnings

# checks lists shamelessly copied from
# https://github.com/koaning/human-learn/blob/master/tests/conftest.py
classifier_checks = (
    estimator_checks.check_classifier_data_not_an_array,
    estimator_checks.check_classifiers_one_label,
    estimator_checks.check_classifiers_classes,
    estimator_checks.check_estimators_partial_fit_n_features,
    estimator_checks.check_classifiers_train,
    estimator_checks.check_supervised_y_2d,
    estimator_checks.check_supervised_y_no_nan,
    estimator_checks.check_estimators_unfitted,
    estimator_checks.check_non_transformer_estimators_n_iter,
    estimator_checks.check_decision_proba_consistency,
)

transformer_checks = (
    estimator_checks.check_transformer_data_not_an_array,
    estimator_checks.check_transformer_general,
    estimator_checks.check_transformers_unfitted,
)

general_checks = (
    estimator_checks.check_fit2d_predict1d,
    estimator_checks.check_methods_subset_invariance,
    estimator_checks.check_fit2d_1sample,
    estimator_checks.check_fit2d_1feature,
    estimator_checks.check_fit1d,
    estimator_checks.check_get_params_invariance,
    estimator_checks.check_set_params,
    estimator_checks.check_dict_unchanged,
    estimator_checks.check_dont_overwrite_parameters,
)

nonmeta_checks = (
    estimator_checks.check_estimators_pickle,
    estimator_checks.check_estimators_dtypes,
    estimator_checks.check_fit_score_takes_y,
    estimator_checks.check_dtype_object,
    estimator_checks.check_estimators_fit_returns_self,
    estimator_checks.check_complex_data,
    estimator_checks.check_estimators_empty_data_messages,
    estimator_checks.check_pipeline_consistency,
    estimator_checks.check_estimators_nan_inf,
    estimator_checks.check_estimators_overwrite_params,
    estimator_checks.check_estimator_sparse_matrix,
    estimator_checks.check_estimator_sparse_array,
)


def select_tests(include, exclude=[]):
    """Return an iterable of include with all tests whose name is not in exclude.

    Credits: https://github.com/koaning/human-learn/blob/master/tests/conftest.py
    """
    for test in include:
        if test.__name__ not in exclude:
            yield test


def flatten(nested_iterable):
    """
    Returns an iterator of flattened values from an arbitrarily nested iterable.

    Usage:

    ```python
    from hulearn.common import flatten
    res1 = list(flatten([['test1', 'test2'], ['a', 'b', ['c', 'd']]]))
    res2 = list(flatten(['test1', ['test2']]))
    assert res1 == ['test1', 'test2', 'a', 'b', 'c', 'd']
    assert res2 == ['test1', 'test2']
    ```

    Credits: https://github.com/koaning/human-learn/blob/master/hulearn/common.py
    """  # noqa
    for el in nested_iterable:
        if isinstance(el, collections.abc.Iterable) and not isinstance(el, (str, bytes)):
            yield from flatten(el)
        else:
            yield el


@pytest.mark.parametrize(
    "transformer,test_fn",
    list(
        itertools.product(
            TRANSFORMERS,
            select_tests(
                include=flatten([general_checks, transformer_checks, nonmeta_checks]),
                exclude=[
                    "check_fit2d_1sample",
                    "check_methods_subset_invariance",
                    "check_estimators_nan_inf",
                    "check_estimator_sparse_matrix",
                    "check_estimator_sparse_array",
                    "check_estimators_empty_data_messages",
                    "check_transformer_data_not_an_array",
                    "check_dtype_object",
                    "check_complex_data",
                    "check_fit1d",
                    "check_transformers_unfitted",
                ],
            ),
        )
    ),
)
def test_transformer_checks(transformer, test_fn):
    """
    Runs a scikitlearn check on a skorecard transformer.
    """
    t = transformer()
    test_fn(t.__class__.__name__, t)


@pytest.mark.parametrize(
    "classifier,test_fn",
    list(
        itertools.product(
            CLASSIFIERS,
            select_tests(
                include=flatten([general_checks, classifier_checks, nonmeta_checks]),
                exclude=[
                    "check_methods_subset_invariance",
                    "check_fit2d_1sample",
                    "check_fit2d_1feature",
                    "check_classifier_data_not_an_array",
                    "check_classifiers_one_label",
                    "check_classifiers_classes",
                    "check_classifiers_train",
                    "check_supervised_y_2d",
                    "check_estimators_pickle",
                    "check_pipeline_consistency",
                    "check_fit2d_predict1d",
                    "check_fit1d",
                    "check_dtype_object",
                    "check_complex_data",
                    "check_estimators_empty_data_messages",
                    "check_estimators_nan_inf",
                    "check_supervised_y_no_nan",
                    "check_estimators_partial_fit_n_features",
                ],
            ),
        )
    ),
)
def test_classifier_checks(classifier, test_fn):
    """
    Runs a scikitlearn check on a skorecard transformer.
    """
    clf = classifier()
    test_fn(clf.__class__.__name__, clf)


def test_cross_val(df):
    """
    Test using CV.

    When defining specials combined with using CV, we would get a

    ValueError: Specials should be defined on the BucketingProcess level,
    remove the specials from DecisionTreeBucketer(specials={'EDUCATION': {'Some specials': [1, 2]}})

    This unit test ensures specials with CV keep working.
    """
    y = df["default"].values
    X = df.drop(columns=["default", "pet_ownership"])

    specials = {"EDUCATION": {"Some specials": [1, 2]}}

    bucketing_process = BucketingProcess(
        prebucketing_pipeline=make_pipeline(
            DecisionTreeBucketer(max_n_bins=100, min_bin_size=0.05),
        ),
        bucketing_pipeline=make_pipeline(
            OptimalBucketer(max_n_bins=10, min_bin_size=0.05),
        ),
        specials=specials,
    )

    pipe = make_pipeline(bucketing_process, StandardScaler(), LogisticRegression(solver="liblinear", random_state=0))

    cross_val_score(pipe, X, y, cv=5, scoring="roc_auc")


def test_cv_pipeline(df):
    """
    Another CV.
    """
    y = df["default"].values
    X = df.drop(columns=["default", "pet_ownership"])

    specials = {"EDUCATION": {"Some specials": [1, 2]}}

    bucketing_process = BucketingProcess(
        prebucketing_pipeline=make_pipeline(
            DecisionTreeBucketer(max_n_bins=100, min_bin_size=0.05),
        ),
        bucketing_pipeline=make_pipeline(
            OptimalBucketer(max_n_bins=10, min_bin_size=0.05),
        ),
        specials=specials,
    )

    pipe = make_pipeline(
        bucketing_process, WOEEncoder(cols=X.columns), LogisticRegression(solver="liblinear", random_state=0)
    )

    with warnings.catch_warnings(record=True) as caught_warnings:
        warnings.simplefilter("always")
        cross_val_score(pipe, X, y, cv=5, scoring="roc_auc")
        return len(caught_warnings) == 0
